@@ -17,6 +17,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <string.h>
 #include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
 #include "gailbutton.h"
@@ -32,8 +33,8 @@ static AtkStateSet*          gail_button_ref_state_set    (AtkObject       *obj)
 static void                  gail_button_real_notify_gtk  (GObject         *obj,
                                                            GParamSpec      *pspec);
 
-static void                  gail_button_real_init        (GailWidget      *widget,
-                                                           GtkWidget       *gtk_widget);
+static void                  gail_button_real_initialize  (AtkObject       *obj,
+                                                           gpointer        data);
 static void                  gail_button_finalize         (GObject        *object);
 
 static void                  gail_button_pressed_enter_handler  (GtkWidget       *widget);
@@ -65,7 +66,8 @@ static void                  gail_button_get_image_size (AtkImage       *image,
 static gboolean              gail_button_set_image_description 
                                                         (AtkImage       *image,
                                                          const gchar    *description);
-static GtkImage*             _get_image_from_button     (GtkButton      *button);
+static GtkImage*             get_image_from_button      (GtkWidget      *button);
+static GtkWidget*            get_label_from_button      (GtkWidget      *button);
 
 static GailContainer* parent_class = NULL;
 
@@ -75,43 +77,43 @@ gail_button_get_type (void)
   static GType type = 0;
 
   if (!type)
-  {
-    static const GTypeInfo tinfo =
     {
-      sizeof (GailButtonClass),
-      (GBaseInitFunc) NULL, /* base init */
-      (GBaseFinalizeFunc) NULL, /* base finalize */
-      (GClassInitFunc) gail_button_class_init, /* class init */
-      (GClassFinalizeFunc) NULL, /* class finalize */
-      NULL, /* class data */
-      sizeof (GailButton), /* instance size */
-      0, /* nb preallocs */
-      (GInstanceInitFunc) gail_button_object_init, /* instance init */
-      NULL /* value table */
-    };
+      static const GTypeInfo tinfo =
+      {
+        sizeof (GailButtonClass),
+        (GBaseInitFunc) NULL, /* base init */
+        (GBaseFinalizeFunc) NULL, /* base finalize */
+        (GClassInitFunc) gail_button_class_init, /* class init */
+        (GClassFinalizeFunc) NULL, /* class finalize */
+        NULL, /* class data */
+        sizeof (GailButton), /* instance size */
+        0, /* nb preallocs */
+        (GInstanceInitFunc) gail_button_object_init, /* instance init */
+        NULL /* value table */
+      };
 
-    static const GInterfaceInfo atk_action_info =
-    {
+      static const GInterfaceInfo atk_action_info =
+      {
         (GInterfaceInitFunc) atk_action_interface_init,
         (GInterfaceFinalizeFunc) NULL,
         NULL
-    };
+      };
 
-	static const GInterfaceInfo atk_image_info =
-    {
+      static const GInterfaceInfo atk_image_info =
+      {
         (GInterfaceInitFunc) atk_image_interface_init,
         (GInterfaceFinalizeFunc) NULL,
         NULL
-    };
+      };
 
-    type = g_type_register_static (GAIL_TYPE_CONTAINER,
-                                   "GailButton", &tinfo, 0);
+      type = g_type_register_static (GAIL_TYPE_CONTAINER,
+                                     "GailButton", &tinfo, 0);
 
-    g_type_add_interface_static (type, ATK_TYPE_ACTION,
-                                 &atk_action_info);
-    g_type_add_interface_static (type, ATK_TYPE_IMAGE,
-                                 &atk_image_info);
-  }
+      g_type_add_interface_static (type, ATK_TYPE_ACTION,
+                                   &atk_action_info);
+      g_type_add_interface_static (type, ATK_TYPE_IMAGE,
+                                   &atk_image_info);
+    }
 
   return type;
 }
@@ -127,14 +129,14 @@ gail_button_class_init (GailButtonClass *klass)
 
   gobject_class->finalize = gail_button_finalize;
 
-  parent_class = g_type_class_ref (GAIL_TYPE_CONTAINER);
+  parent_class = g_type_class_peek_parent (klass);
 
   class->get_name = gail_button_get_name;
   class->get_n_children = gail_button_get_n_children;
   class->ref_child = gail_button_ref_child;
   class->ref_state_set = gail_button_ref_state_set;
+  class->initialize = gail_button_real_initialize;
 
-  widget_class->init = gail_button_real_init;
   widget_class->notify_gtk = gail_button_real_notify_gtk;
 }
 
@@ -147,22 +149,20 @@ gail_button_object_init (GailButton      *button)
   button->click_keybinding = NULL;
 }
 
-GtkAccessible* 
+AtkObject* 
 gail_button_new (GtkWidget *widget)
 {
   GObject *object;
-  GtkAccessible *accessible;
+  AtkObject *accessible;
 
   g_return_val_if_fail (GTK_IS_BUTTON (widget), NULL);
 
   object = g_object_new (GAIL_TYPE_BUTTON, NULL);
 
-  g_return_val_if_fail (GTK_IS_ACCESSIBLE (object), NULL);
+  accessible = ATK_OBJECT (object);
+  atk_object_initialize (accessible, widget);
 
-  gail_widget_init (GAIL_WIDGET (object), widget);
-
-  accessible = GTK_ACCESSIBLE (object);
-  ATK_OBJECT(accessible)->role = ATK_ROLE_PUSH_BUTTON;
+  accessible->role = ATK_ROLE_PUSH_BUTTON;
 
   return accessible;
 }
@@ -173,80 +173,64 @@ gail_button_get_name (AtkObject *obj)
   g_return_val_if_fail (GAIL_IS_BUTTON (obj), NULL);
 
   if (obj->name != NULL)
-  {
     return obj->name;
-  }
   else
-  {
-    /*
-     * Get the text on the label
-     */
-    GtkWidget *widget;
-    GtkWidget *child;
-    G_CONST_RETURN gchar* name = NULL;
-
-    widget = GTK_ACCESSIBLE (obj)->widget;
-    if (widget == NULL)
     {
       /*
-       * State is defunct
+       * Get the text on the label
        */
-      return NULL;
-    }
-    g_return_val_if_fail (GTK_IS_BUTTON (widget), NULL);
+      GtkWidget *widget;
+      GtkWidget *child;
+      G_CONST_RETURN gchar* name = NULL;
 
-    child = gtk_bin_get_child (GTK_BIN (widget));
-    if (GTK_IS_BOX (child))
-    {
-      /*
-       * Child is not a label; perhaps the button is for a stock item.
-       * If this is so, then the second child of the child should be a label.
-       */
-      GList *children, *tmp_list;
+      widget = GTK_ACCESSIBLE (obj)->widget;
+      if (widget == NULL)
+        /*
+         * State is defunct
+         */
+        return NULL;
 
-      children = gtk_container_children (GTK_CONTAINER (child));
-      tmp_list = g_list_nth (children, 1);
-      child = GTK_WIDGET (tmp_list->data);
-      g_list_free (children);
+      g_return_val_if_fail (GTK_IS_BUTTON (widget), NULL);
+
+      child = get_label_from_button (widget);
+      if (GTK_IS_LABEL (child))
+        name = gtk_label_get_text (GTK_LABEL (child)); 
+
+      return name;
     }
-    if (GTK_IS_LABEL (child))
-    {
-      name = gtk_label_get_text (GTK_LABEL (child)); 
-    }
-    return name;
-  }
 }
 
 static void
-gail_button_real_init (GailWidget         *widget,
-                       GtkWidget          *gtk_widget)
+gail_button_real_initialize (AtkObject *obj,
+                             gpointer   data)
 {
-  GailContainer *container = GAIL_CONTAINER (widget);
-  GailButton *button = GAIL_BUTTON (widget);
-  GAIL_WIDGET_CLASS (parent_class)->init (widget, gtk_widget);
+  GailContainer *container = GAIL_CONTAINER (obj);
+  GailButton *button = GAIL_BUTTON (obj);
+
+  ATK_OBJECT_CLASS (parent_class)->initialize (obj, data);
 
   /*
    * As we report the button as having no children we are not interested
    * in add and remove signals
    */
-  g_signal_handler_disconnect (G_OBJECT (gtk_widget), container->add_handler);
-  g_signal_handler_disconnect (G_OBJECT (gtk_widget), container->remove_handler);
+  g_signal_handler_disconnect (data, container->add_handler);
+  g_signal_handler_disconnect (data, container->remove_handler);
 
   button->state = GTK_STATE_NORMAL;
 
-  g_signal_connect (G_OBJECT (gtk_widget),
+  g_signal_connect (data,
                     "pressed",
                     G_CALLBACK (gail_button_pressed_enter_handler),
                     NULL);
-  g_signal_connect (G_OBJECT (gtk_widget),
+  g_signal_connect (data,
                     "enter",
                     G_CALLBACK (gail_button_pressed_enter_handler),
                     NULL);
-  g_signal_connect (G_OBJECT (gtk_widget),
+  g_signal_connect (data,
                     "released",
                     G_CALLBACK (gail_button_released_leave_handler),
                     NULL);
-  g_signal_connect (G_OBJECT (gtk_widget),
+  g_signal_connect (data,
                     "leave",
                     G_CALLBACK (gail_button_released_leave_handler),
                     NULL);
@@ -260,28 +244,23 @@ gail_button_real_notify_gtk (GObject           *obj,
   AtkObject* atk_obj = gtk_widget_get_accessible (widget);
 
   if (strcmp (pspec->name, "label") == 0)
-  {
-    g_object_freeze_notify (G_OBJECT (atk_obj));
-    if (atk_obj->name == NULL)
     {
+      if (atk_obj->name == NULL)
+      {
+        /*
+         * The label has changed so notify a change in accessible-name
+         */
+        g_object_notify (G_OBJECT (atk_obj), "accessible-name");
+      }
       /*
-       * The label has changed so notify a change in accessible-name
+       * The label is the only property which can be changed
        */
-      g_object_notify (G_OBJECT (atk_obj), "accessible-name");
+      g_signal_emit_by_name (atk_obj, "visible_data_changed");
     }
-    /*
-     * The label is the only property which can be changed
-     */
-    g_object_notify (G_OBJECT (atk_obj), "accessible-visible-data");
-    g_object_thaw_notify (G_OBJECT (atk_obj));
-  }
   else
-  {
     GAIL_WIDGET_CLASS (parent_class)->notify_gtk (obj, pspec);
-  }
 }
 
-
 static void
 atk_action_interface_init (AtkActionIface *iface)
 {
@@ -305,39 +284,37 @@ gail_button_do_action (AtkAction *action,
 
   widget = GTK_ACCESSIBLE (action)->widget;
   if (widget == NULL)
-  {
     /*
      * State is defunct
      */
     return FALSE;
-  }
 
   button = GTK_BUTTON (widget); 
   switch (i)
-  {
+    {
     case 0:
       gtk_widget_activate (widget);
       break;
     case 1:
-    {
-      GdkEvent tmp_event;
+      {
+        GdkEvent tmp_event;
 
-      button->in_button = TRUE;
-      gtk_button_enter (button);
-      /*
-       * Simulate a button press event. calling gtk_button_pressed() does
-       * not get the job done for a GtkOptionMenu.  
-       */
-      tmp_event.button.type = GDK_BUTTON_PRESS;
-      tmp_event.button.window = widget->window;
-      tmp_event.button.button = 1;
-      tmp_event.button.send_event = TRUE;
-      tmp_event.button.time = GDK_CURRENT_TIME;
-      tmp_event.button.axes = NULL;
+        button->in_button = TRUE;
+        gtk_button_enter (button);
+        /*
+         * Simulate a button press event. calling gtk_button_pressed() does
+         * not get the job done for a GtkOptionMenu.  
+         */
+        tmp_event.button.type = GDK_BUTTON_PRESS;
+        tmp_event.button.window = widget->window;
+        tmp_event.button.button = 1;
+        tmp_event.button.send_event = TRUE;
+        tmp_event.button.time = GDK_CURRENT_TIME;
+        tmp_event.button.axes = NULL;
 
-      gtk_widget_event (widget, &tmp_event);
-      break;
-    }
+        gtk_widget_event (widget, &tmp_event);
+        break;
+      }
     case 2:
       button->in_button = FALSE;
       gtk_button_leave (button);
@@ -346,7 +323,7 @@ gail_button_do_action (AtkAction *action,
     default:
       return_value = FALSE;
       break;
-  }
+    }
   return return_value; 
 }
 
@@ -366,7 +343,7 @@ gail_button_get_description (AtkAction *action,
   button = GAIL_BUTTON (action);
 
   switch (i)
-  {
+    {
     case 0:
       return_value = button->click_description;
       break;
@@ -379,7 +356,7 @@ gail_button_get_description (AtkAction *action,
     default:
       return_value = NULL;
       break;
-  }
+    }
   return return_value; 
 }
 
@@ -392,55 +369,41 @@ gail_button_get_keybinding (AtkAction *action,
 
 
   switch (i)
-  {
-    case 0:
     {
-      /*
-       * We look for a mnemonic on the label
-       */
-      GtkWidget *widget;
-      GtkWidget *child;
-
-      button = GAIL_BUTTON (action);
-      widget = GTK_ACCESSIBLE (button)->widget;
-      if (widget == NULL)
+    case 0:
       {
         /*
-         * State is defunct
+         * We look for a mnemonic on the label
          */
-        return NULL;
-      }
-      g_return_val_if_fail (GTK_IS_BUTTON (widget), NULL);
+        GtkWidget *widget;
+        GtkWidget *child;
 
-      child = gtk_bin_get_child (GTK_BIN (widget));
-      if (GTK_IS_BOX (child))
-      {
-        /*
-         * Child is not a label; perhaps the button is for a stock item.
-         * If this is so, then the second child of the child should be a label.
-         */
-        GList *children, *tmp_list;
+        button = GAIL_BUTTON (action);
+        widget = GTK_ACCESSIBLE (button)->widget;
+        if (widget == NULL)
+          /*
+           * State is defunct
+           */
+          return NULL;
 
-        children = gtk_container_children (GTK_CONTAINER (child));
-        tmp_list = g_list_nth (children, 1);
-        child = GTK_WIDGET (tmp_list->data);
-        g_list_free (children);
-      }
-      if (GTK_IS_LABEL (child))
-      {
-        guint key_val; 
+        g_return_val_if_fail (GTK_IS_BUTTON (widget), NULL);
 
-        key_val = gtk_label_get_mnemonic_keyval (GTK_LABEL (child)); 
-        if (key_val != GDK_VoidSymbol)
-          return_value = gtk_accelerator_name (key_val, 0);
+        child = get_label_from_button (widget);
+        if (GTK_IS_LABEL (child))
+          {
+            guint key_val; 
+
+            key_val = gtk_label_get_mnemonic_keyval (GTK_LABEL (child)); 
+            if (key_val != GDK_VoidSymbol)
+              return_value = gtk_accelerator_name (key_val, 0);
+            g_free (button->click_keybinding);
+            button->click_keybinding = g_strdup (return_value);
+          }
+        break;
       }
-      g_free (button->click_keybinding);
-      button->click_keybinding = return_value;
-      break;
-    }
     default:
       break;
-  }
+    }
   return return_value; 
 }
 
@@ -451,7 +414,7 @@ gail_button_action_get_name (AtkAction *action,
   G_CONST_RETURN gchar *return_value;
 
   switch (i)
-  {
+    {
     case 0:
       /*
        * This action is a "click" to activate a button or "toggle" to change
@@ -479,7 +442,7 @@ gail_button_action_get_name (AtkAction *action,
     default:
       return_value = NULL;
       break;
-  }
+    }
   return return_value; 
 }
 
@@ -494,7 +457,7 @@ gail_button_set_description (AtkAction      *action,
   button = GAIL_BUTTON (action);
 
   switch (i)
-  {
+    {
     case 0:
       value = &button->click_description;
       break;
@@ -507,17 +470,15 @@ gail_button_set_description (AtkAction      *action,
     default:
       value = NULL;
       break;
-  }
+    }
   if (value)
-  {
-    g_free (*value);
-    *value = g_strdup (desc);
-    return TRUE;
-  }
+    {
+      g_free (*value);
+      *value = g_strdup (desc);
+      return TRUE;
+    }
   else
-  {
     return FALSE;
-  }
 }
 
 /*
@@ -568,20 +529,13 @@ static void
 gail_button_pressed_enter_handler (GtkWidget       *widget)
 {
   AtkObject *accessible;
-  AtkPropertyValues  values;
-
-  memset (&values.old_value, 0, sizeof (GValue));
-  memset (&values.new_value, 0, sizeof (GValue));
-  values.property_name = "accessible-state";
 
   if (GTK_WIDGET_STATE (widget) == GTK_STATE_ACTIVE)
-  {
-    g_value_init (&values.new_value, G_TYPE_INT);
-    g_value_set_int (&values.new_value, ATK_STATE_ARMED);
-    accessible = gtk_widget_get_accessible (widget);
-    g_signal_emit_by_name (accessible,  "property-change", &values, NULL);
-    GAIL_BUTTON (accessible)->state = GTK_STATE_ACTIVE;
-  }
+    {
+      accessible = gtk_widget_get_accessible (widget);
+      atk_object_notify_state_change (accessible, ATK_STATE_ARMED, TRUE);
+      GAIL_BUTTON (accessible)->state = GTK_STATE_ACTIVE;
+    }
 }
 
 /*
@@ -594,20 +548,13 @@ static void
 gail_button_released_leave_handler (GtkWidget       *widget)
 {
   AtkObject *accessible;
-  AtkPropertyValues  values;
-
-  memset (&values.old_value, 0, sizeof (GValue));
-  memset (&values.new_value, 0, sizeof (GValue));
-  values.property_name = "accessible-state";
 
   accessible = gtk_widget_get_accessible (widget);
   if (GAIL_BUTTON (accessible)->state == GTK_STATE_ACTIVE)
-  {
-    g_value_init (&values.old_value, G_TYPE_INT);
-    g_value_set_int (&values.old_value, ATK_STATE_ARMED);
-    g_signal_emit_by_name (accessible,  "property-change", &values, NULL);
-    GAIL_BUTTON (accessible)->state = GTK_STATE_NORMAL;
-  }
+    {
+      atk_object_notify_state_change (accessible, ATK_STATE_ARMED, FALSE);
+      GAIL_BUTTON (accessible)->state = GTK_STATE_NORMAL;
+    }
 }
 
 static void
@@ -621,22 +568,24 @@ atk_image_interface_init (AtkImageIface *iface)
 }
 
 static GtkImage*
-_get_image_from_button (GtkButton *button)
+get_image_from_button (GtkWidget *button)
 {
   GtkWidget *child;
-  GList 	*list;
+  GList *list;
+  GtkImage *image = NULL;
 
   child = gtk_bin_get_child (GTK_BIN (button));
+  if (GTK_IS_ALIGNMENT (child))
+    child = gtk_bin_get_child (GTK_BIN (child));
   if (GTK_IS_BOX (child))
-  {
-    list = gtk_container_children (GTK_CONTAINER (child));
-    if (GTK_IS_IMAGE (list->data))
     {
-      return GTK_IMAGE (list->data);
+      list = gtk_container_children (GTK_CONTAINER (child));
+      if (GTK_IS_IMAGE (list->data))
+        image = GTK_IMAGE (list->data);
+      g_list_free (list);
     }
-  }
 
-  return NULL;
+  return image;
 }
 
 static G_CONST_RETURN gchar* 
@@ -649,19 +598,18 @@ gail_button_get_image_description (AtkImage *image) {
   widget = GTK_ACCESSIBLE (image)->widget;
 
   if (widget == NULL)
-  {
     /*
      * State is defunct
      */
     return NULL;
-  }
 
-  button_image = _get_image_from_button(GTK_BUTTON(widget));
+  button_image = get_image_from_button (widget);
 
-  if (button_image != NULL) {
-    obj = gtk_widget_get_accessible(GTK_WIDGET(button_image));
-    return atk_image_get_image_description(ATK_IMAGE(obj));
-  }
+  if (button_image != NULL)
+    {
+      obj = gtk_widget_get_accessible (GTK_WIDGET (button_image));
+      return atk_image_get_image_description (ATK_IMAGE (obj));
+    }
   else 
     return NULL;
 }
@@ -678,27 +626,25 @@ gail_button_get_image_size (AtkImage *image,
   widget = GTK_ACCESSIBLE (image)->widget;
 
   if (widget == NULL)
-  {
     /*
      * State is defunct
      */
     *width = -1;
     *height = -1;
     return;
-  }
 
-  button_image = _get_image_from_button (GTK_BUTTON (widget));
+  button_image = get_image_from_button (widget);
 
   if (button_image != NULL)
-  {
-    obj = gtk_widget_get_accessible (GTK_WIDGET (button_image));
-    atk_image_get_image_size (ATK_IMAGE (obj), width, height); 
-  }
+    {
+      obj = gtk_widget_get_accessible (GTK_WIDGET (button_image));
+      atk_image_get_image_size (ATK_IMAGE (obj), width, height); 
+    }
   else
-  {
-    *width = -1;
-    *height = -1;
-  }
+    {
+      *width = -1;
+      *height = -1;
+    }
 }
 
 static gboolean
@@ -712,21 +658,18 @@ gail_button_set_image_description (AtkImage    *image,
   widget = GTK_ACCESSIBLE (image)->widget;
 
   if (widget == NULL)
-  {
     /*
      * State is defunct
      */
     return FALSE;
-  }
-
-  button_image = _get_image_from_button (GTK_BUTTON (widget));
 
+  button_image = get_image_from_button (widget);
 
   if (button_image != NULL) 
-  {
-    obj = gtk_widget_get_accessible (GTK_WIDGET (button_image));
-    return atk_image_set_image_description (ATK_IMAGE (obj),description);
-  }
+    {
+      obj = gtk_widget_get_accessible (GTK_WIDGET (button_image));
+      return atk_image_set_image_description (ATK_IMAGE (obj), description);
+    }
   else 
     return FALSE;
 }
@@ -742,3 +685,27 @@ gail_button_finalize (GObject            *object)
   g_free (button->click_keybinding);
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
+
+static GtkWidget*
+get_label_from_button (GtkWidget *button)
+{
+  GtkWidget *child;
+
+  child = gtk_bin_get_child (GTK_BIN (button));
+  if (GTK_IS_ALIGNMENT (child))
+    child = gtk_bin_get_child (GTK_BIN (child));
+  if (GTK_IS_BOX (child))
+    {
+      /*
+       * Child is not a label; perhaps the button is for a stock item.
+       * If it is, then the second child of the child should be a label.
+       */
+      GList *children, *tmp_list;
+ 
+      children = gtk_container_children (GTK_CONTAINER (child));
+      tmp_list = g_list_nth (children, 1);
+      child = GTK_WIDGET (tmp_list->data);
+      g_list_free (children);
+    }
+  return child;
+}
