@@ -27,7 +27,7 @@ static void gail_paned_factory_class_init (GailPanedFactoryClass       *klass);
 static AtkObject* gail_paned_factory_create_accessible (
                                GObject                              *obj);
 
-static AtkObjectFactoryClass    *parent_class = NULL;
+static GType gail_paned_factory_get_accessible_type (void);
 
 GType
 gail_paned_factory_get_type ()
@@ -61,12 +61,12 @@ gail_paned_factory_class_init (GailPanedFactoryClass *klass)
 {
   AtkObjectFactoryClass *class = ATK_OBJECT_FACTORY_CLASS (klass);
 
-  parent_class = g_type_class_ref (ATK_TYPE_OBJECT_FACTORY);
-
   class->create_accessible = gail_paned_factory_create_accessible;
+  class->get_accessible_type = gail_paned_factory_get_accessible_type;
 }
 
-AtkObjectFactory* gail_paned_factory_new ()
+AtkObjectFactory*
+gail_paned_factory_new ()
 {
   GObject *factory;
 
@@ -86,3 +86,9 @@ gail_paned_factory_create_accessible (GObject     *obj)
 
   return gail_paned_new (widget);
 }
+
+static GType
+gail_paned_factory_get_accessible_type (void)
+{
+  return GAIL_TYPE_PANED;
+}
